/*

Read a symbol table file and sort it.
Input from one or more named files, output to stdout.  

The output is one row per file, with the name of the file, the number
of symbols in the file, followed by the sort times for each of the
sorters.  The file is in CSV (comma separated variables) format,
suitable for importing into Excel or other spreadsheet programs.

Doug Johnson, djohnson@cs.washington.edu, for CSE 373, May 2002.

The format of each row of the input symbol table file is simply

  <numeric value> <symbol name>

See symbol.h and symbol.c for more information about Symbols.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <limits.h>

#include "reporter.h"
#include "sort.h"
#include "symbol.h"

int checkSortByName(ElementType *a, int count);
Sorter NullSort;

struct SortListItem {
  char *label;			/* label for this sorter */
  Sorter *S;			/* the actual sort function */
  int maxCount;			/* max number of items to apply sorter to */
};

struct SortListItem SortList[] = {
  {"Insert",InsertionSort,1000},
  {"Shell",ShellSort,1000000},
  {"Heap",HeapSort,1000000},
  {"Merge",MergeSort,1000000},
  {"Quick",QuickSort,1000000},
  {NULL, NULL, 0}
};

int main(int argc,char *argv[]) {
  FILE *src;			/* input FILE */

  ElementType *A;		/* element pointers - sorted */
  ElementType *B;		/* element pointers - as read from file */

  int idx;			/* loop control */
  int count;			/* number of symbols provided */
  int errors;			/* sort error count */
  int arg;			/* current filename argument */
  clock_t start;		/* timer values */
  clock_t elapsed;
  clock_t duration;

  int nullPassCount;		/* number of passes in overhead test */
  int passCount;		/* number of passes in current test */
  int maxPassCount;		/* maximum number of passes for each test */
  int maxDuration;		/* maximum number of seconds for each test */

  maxPassCount = 100000;
  maxDuration = 5*CLOCKS_PER_SEC;

  printf("Count");
  idx = 0;
  while (SortList[idx].label != NULL) {
    printf(",%s",SortList[idx].label);
    idx++;
  }
  printf(",Filename\n");

  /* read from the named file(s) */

  arg = 0;
  while (++arg < argc) {
    src = fopen(argv[arg],"r");
    if (src == NULL) {
      printf("%s : file open error.\n",argv[arg]);
      break;
    }

    /* Read the symbol definitions. */

    B = (ElementType *)readSymbols(src);
    fclose(src);

    /* Count Symbols */

    count = 0;
    while (B[count] != NULL) {
      count++;
    }
    printf("%i",count);

    /* Make a copy for sorting */

    A = malloc(count*sizeof(ElementType)); /* array to sort */
    FatalErrorMemory(A);

    /* Use each defined sorter to massage the data */

    idx = 0;
    while (SortList[idx].label != NULL) {
      if (count <= SortList[idx].maxCount) {

	/* run a test with the specified sorter */

	passCount = 0;
	elapsed = 0;
	start = clock();
	while (elapsed < maxDuration && passCount < maxPassCount) {
	  memcpy(A,B,count*sizeof(ElementType));
	  (SortList[idx].S)(A,count,symbolCompareByName);
	  elapsed = clock()-start;
	  passCount++;
	}
	duration = elapsed;
	errors = checkSortByName(A,count);

	/* estimate how much of that was overhead time */

	nullPassCount = 0;
	elapsed = 0;
	start = clock();
	while (elapsed < LONG_MAX && nullPassCount < passCount) {
	  memcpy(A,B,count*sizeof(ElementType));
	  NullSort(A,count,symbolCompareByName);
	  elapsed = clock()-start;
	  nullPassCount++;
	}

	/* actual duration was time not spent in overhead work */

	duration = duration - elapsed;
	if (duration >= elapsed) /* make sure overhead is reasonable */
	  duration = duration - elapsed;
	else
	  duration = 0;
	  
	if (errors == 0) {	/* sort worked - print seconds/pass */
	  printf(",%7.5f",(double)duration/(double)passCount/CLOCKS_PER_SEC);
	} else {
	  printf(",%i",-errors); /* sort did not work correctly */
	}
      } else {
	printf(",");		/* test not run - data set too large */
      }
      idx++;
    }
    printf(",%s\n",argv[arg]);
  }
  return 0;
}

int checkSortByName(ElementType *a, int count) {
  int i;
  int error;
  error = 0;
  for (i=1; i<count; i++) {
    if (symbolCompareByName(a[i-1],a[i])>0) error++;
  }
  return error;
}

void NullSort(ElementType A[],int count, Comparator C){}
